/**********************************************************************
 *
 *       moonsun.c  -  Routines to find Moon and Sun data.
 *                        Version 1.9
 *
 * NOTICE: Copyright (C) 1996, 1997, 1998, 1999 by Robert S. Larkin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The program author can be contacted by email at boblark@proaxis.com.
 * Alternatively contact the author at
 *     Robert S. (Bob) Larkin
 *     2982 N. W. Acacia Place
 *     Corvallis, OR  97330  USA
 *
 * For more details on the background or use of the program, pay a visit to
 * http://www.proaxis.com/~boblark/dsp10.htm 
 *
 * The associated programs are based on a variety of sources including
 * the Astronomical Almanac and its background book. Also the Naval
 * Observatory web site and discussions of time scales. Major credit must
 * go to the several books for there algorithmic contributions:
 *  Peter Duffett-Smith, "Astronomy with your Personal Computer," Second
 *        edition, Cambridge University Press, 1990.
 *  Jean Meeus, "Astronomical Formulae for Calculators," Willman-Bell, Inc,.
 *        1982, 4th ed.
 *  Oliver Montenbruck, "Practical Ephemeris Calculations," Springer-Verlag,
 *        1987.
 *  U.S. Naval Observatory, "Multiyear Interactive Computer Almanac,"
 *        Willman-Bell, Inc., 1998.
 *
 *
 * Initial write 17 Jan 1999 RSL
 * 9 Feb 98  Convert so all angles, except for display, are in radians.  
 * 29 Mar 99 Remove test drivers to make into subroutine library.  RSL
 * 11 April 01  Completed commenting of this file
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>
#include <string.h>
#include <time.h>
#include "moonsun.h"
#include "uhfa.h"
/* For diagnostic print: */
#include <dos.h>
#include <graphics.h>

/* doppler()   Returns doppler shift in Hz for EM path (multiply by 2 for
 * own echos, or add two shifts for full path.)
 * Inputs:
 *      *ptm structure of type tm with current UTC
 *      ddj = Julian day
 *	plc->lon = local geographical longitude (radians)
 *	plc->lat = local geographical latitude (radians)
 *	fhz = frequency (Hz )
 *
 * Outputs:
 *      Returned Doppler shift in Hz
 *	pae->el = local elevation angle of Moon (radians)
 *	pae->az = local azimuth of Moon (radians)
 *      *prem = range E-M in meters for loc[0]
 *      
 *	ut = Universal Time (global variable), filled in from *ptm
 */
double doppler(struct tm *ptm, struct azel *pae, double *prem,
		 double fhz, struct location *plc)
    {
    #define K0  0.0727220521664	/* 1000.0*TPI/(24.0*3600.0) */
    #define DELTATIME 1.0	/* For finding derivatives (seconds) */

    struct radec mrd0, mrd1;		/* Storage for RA and Dec */
    struct hadec mhd0, mhd1;		/* Hour angle and declination */
    double pm0, pm1;			/* Moon horizontal parallax */
    double rem0, rem1;			/* Distance E-M centers, m */
    double rd0, rd1;			/* 2 Ranges, for Doppler Calc */
    double tp0, tp1, tp2;		/* Temporary working variables */
    double xm, ym, zm;			/* Moon, All in meters */
    double xo, zo;			/* Observer */
    double dltx, dltz, lt, lta;
    double gst;
    struct latlon mll0, mll1;		/* Ecliptic lat, lon of Moon */

    xo=REARTH *cos(plc->lat);		/* Ellipsoidal Earth */
    zo=REARTHP*sin(plc->lat);
    lt=r2d(plc->lat);
    lta=lt;
    dltx=-470.28+95.529*lta;
    dltz=671.12-969.19*lta;
    lta *= lt;
    dltx += 4.3823*lta;
    dltz += 12.343*lta;
    lta *= lt;
    dltx -= 0.0608*lta;
    dltz -= 0.0471*lta;
    xo += dltx;
    zo += dltz;

    /* cnvlt2gst() will fill in value for ut (global) for use by moon() */
    gst=cnvlt2gst( (double)ptm->tm_hour +
	((((double)ptm->tm_min) + ((double)ptm->tm_sec/60.0))/60.0),
	    &cur_env, plc, &ut);

    /* moon() returns the ecliptic coordinates, mll, & pm:  */
    moon(cur_env.dj, &mll0.lon, &mll0.lat, &pm0);

    /* Change pm to range in meters: */
    rem0=REARTH/sin(pm0);

    /* Change ecliptic coordinates to Right Ascension and Declination: */
    cnvll2radec(&mll0, &mrd0, &cur_env);

    /* And change Right Ascension to Local Hour Angle:  */
    mhd0.ha=cnvra2ha(mrd0.ra, gst, plc->lon);
    mhd0.dec=mrd0.dec;			/* Declination is unchanged */

    /* Convert both the observer and Moon coordinates into a Cartesian
     * coordinate system, Earth centered, with the x-z plane going through
     * the Earth's poles and the observers location. The z axis is the
     * Earths polar axis. The observer's y value is 0.0. tp0. tp1, tp2
     * are temporary working variables.
     */

    tp0=cos(mhd0.dec);
    xm=rem0*tp0*cos(mhd0.ha);
    ym=rem0*tp0*sin(mhd0.ha);
    zm=rem0*sin(mhd0.dec);
    tp1=xm-xo;  tp2=zm-zo;

    /* Find the range from the observer to the Moon's center:  */
    rd0=sqrt(tp1*tp1 + ym*ym + tp2*tp2);
    *prem=rd0;				/* Send it back for dB calc */

    /* To find the az and el for the observers location, two coordinate
     * changes are needed. First, move out from the Earth's center:  */
    parallax(plc->lat, &mhd0, pm0, &mhd1);

    /* And then convert from Hour Angle/Declination to Az and El:   */
    cnvhadec2azel(&mhd1, plc, pae);

    /* Compute a second set of values to find range derivative */
    gst=cnvlt2gst( (double)ptm->tm_hour +
	((((double)ptm->tm_min) +
             ((DELTATIME+(double)ptm->tm_sec)/60.0))/60.0),
	     &cur_env, plc, &ut);
    moon(cur_env.dj, &mll1.lon, &mll1.lat, &pm1);
    rem1=REARTH/sin(pm1);
    cnvll2radec(&mll1, &mrd1, &cur_env);
    mhd1.ha=cnvra2ha(mrd1.ra, gst, plc->lon);
    mhd1.dec=mrd1.dec;

    tp0=cos(mhd1.dec);
    xm=rem1*tp0*cos(mhd1.ha);
    ym=rem1*tp0*sin(mhd1.ha);
    zm=rem1*sin(mhd1.dec);
    tp1=xm-xo;  tp2=zm-zo;
    rd1=sqrt(tp1*tp1 + ym*ym + tp2*tp2);

    return fhz*(rd0-rd1)/SOL;
    }

/* Convert decimal degrees into deg, min, sec
 * argument is degrees.  Fills out a structure of type dms.
 * Seconds are rounded to nearest .01 sec.
 */
void cnvdd2dms(double dd, struct dms *padms)
	{
	double x;

	padms->sign=fsign(dd);
	dd=fabs(dd);
	padms->deg=(int)floor(dd);
	dd=dd-floor(dd);
	padms->min=(int)(60.0*dd);
	x=60.0*fract_part(60*dd);
	padms->sec=0.01*floor(100.0*x+0.5);
	if (padms->deg>=100)
	    {
	    sprintf(padms->asc," %.3dd %.2dm %5.2lfs",
		padms->deg, padms->min, padms->sec);
	    if (padms->sign >= 0) padms->asc[0]='+';
	    else   padms->asc[0]='-';
	    }
	else if (padms->deg>=10)
	    {
	    sprintf(padms->asc,"  %.2dd %.2dm %5.2lfs",
		padms->deg, padms->min, padms->sec);
	    if (padms->sign >= 0) padms->asc[1]='+';
	    else   padms->asc[1]='-';
	    }
	else
	    {
	    sprintf(padms->asc,"   %.1dd %.2dm %5.2lfs",
		padms->deg, padms->min, padms->sec);
	    if (padms->sign >= 0) padms->asc[2]='+';
	    else   padms->asc[2]='-';
	    }
	}

/* Convert decimal hours into hours, min, sec
 * argument is hours.  Fills in a structure of type hms.  */
void cnvdh2hms(double dh, struct hms *pthms)
	{
	double x;

	pthms->sign=fsign(dh);
	dh=fabs(dh);
	pthms->hr=(int)floor(dh);
	dh=fract_part(dh);
	pthms->min=(int)(60.0*dh);
	x=60.0*fract_part(60*dh);
	pthms->sec=0.01*floor(100.0*x+0.5);
	if (pthms->hr>=10)
	    {
	    sprintf(pthms->asc," %.2dh %.2dm %5.2lfs",
		pthms->hr, pthms->min, pthms->sec);
	    if (pthms->sign >= 0) pthms->asc[0]='+';
	    else   pthms->asc[0]='-';
	    }
	else
	    {
	    sprintf(pthms->asc,"  %.1dh %.2dm %5.2lfs",
		pthms->hr, pthms->min, pthms->sec);
	    if (pthms->sign >= 0) pthms->asc[1]='+';
	    else   pthms->asc[1]='-';
	    }
	}

/* cnvut2tdt(time_t *put, struct tm *ptdt)
 * Converts domestic ut (WWV) to TDT for ephemeris calculation
 * by adding an offset determined by Bob Larkin curve fit of data as of
 * 1999. Data predicted for 1999 to 2012, from deltat.preds at
 * maia.usno.navy.mil (FTP).
 * 
 */
void cnvut2tdt(time_t *put, struct tm *ptdt)
	{
	struct tm *ptm;
	double y2;
	time_t deltat, ttdt;

deltat=(time_t) near_int(dut1);  /* <<<<<<<<<< TEMP FOR TESTING */

	ttdt = *put + deltat;
	ptm=gmtime(&ttdt);
        /* To cut the risk of other calls to gmtime(), we need to transfer
	 * the data to our own tdt structure:  */
	ptdt->tm_sec  = ptm->tm_sec;
	ptdt->tm_min  = ptm->tm_min;
	ptdt->tm_hour = ptm->tm_hour;
	ptdt->tm_mday = ptm->tm_mday;
	ptdt->tm_mon  = ptm->tm_mon;
	ptdt->tm_year = ptm->tm_year;
	ptdt->tm_wday = ptm->tm_wday;
	ptdt->tm_yday = ptm->tm_yday;
	ptdt->tm_isdst= ptm->tm_isdst;
	}

/* cnvdms2dd(dms*)    Convert deg, min, sec into decimal degrees.
 * Argument is a pointer to a structure of type dms.
 */
double cnvdms2dd(struct dms *pdms)
	{
	return	(double)pdms->sign*
		(((pdms->sec/60.0)+
		(double)pdms->min)/60.0+
		(double)pdms->deg);
	}

/* cnvhms2dh(struct hms *pthms)   Converts a structure of type hms into decimal
 * hour, which is returned.
 */
double cnvhms2dh(struct hms *pthms)
	{
	return	(double)pthms->sign*
		(((pthms->sec/60.0)+
		(double)pthms->min)/60.0+
		(double)pthms->hr);
	}

/* JULDAY - Takes day (1,31), month (1, 12) and year (e.g.1950)
 * Modifies e.dj, the number of Julian days since 1900 January 0.5, and
 * Julian days, jd=dj+2415020.  Returns 0 if succesful and -1
 * if day does not exist.
 */
int julday(struct eenv *e)
	{
	double a, b, c;
	double dy;			/* Local copy of day */
	int    mn, yr;			/* Local copies of month and year */

	dy = e->day;			/* So that we can change them */
	mn = e->month;
	yr = e->year;

	if (yr==0) return -1;		/* No year 0 */
	/* Pope Gregory abolished the following dates: */
	if (yr==1582 && mn==10 && dy>=5.0 && dy<15.0)
		 return -1;
	if (yr<0)  yr++;			/* Remove year 0 */
	if (mn < 3)
	    {
	    mn = mn + 12;
	    yr--;
	    }
	/* Are we using Gregorian calendar?  */
	if (yr>1582 || (yr==1582 && mn>10) ||
		 (yr==1582 && mn==10 && dy>=15))
	    {				/* Yes we are */
	    a=(double)(yr/100);
	    b=2.0-a+floor(a/4.0);
	    }
	else
	    b=0.0;

	if (yr<0)
	    c = floor(365.25*(double)yr - 0.75) - 694025.0;
	else
	    c = floor(365.25*(double)yr) - 694025.0;	
	e->dj = b + c + floor(30.6001*(double)(mn+1)) + dy - 0.5;
	e->jd = e->dj + 2415020.0;
	return 0;
	}

/* Converts local civil time to Greenwich sidereal time in decimal hours,
 * and to *put, Universal time.
 * This differs from Duffett-Smith in that LST is not found.
 * e->dj is an input argument.  A call to julday() should precede time().
 * Inputs:	double time		Local civil time in decimal hours
 *		double e->dj		Julian days since (0.5, 1, 1900)
 *		int e->year		Like 1999
 * 		ploc->time_zone		Hours east of Greenwich
 *		ploc-> dst		0 or 1, 0-->Local time not DST
 * Output:	*put			Universal time in decimal hours
 * Returns:				Greenwich Sidereal Time, dec hours
 */
double cnvlt2gst
    (double time, struct eenv *e, struct location *ploc, double *put)
	{
	int flg;
	double d, t, r0, r1, t0, sg, gst;

	/* First find t0, which is GST at 0h local civil time */	
	d=(floor)(e->dj-0.5) + 0.5;	/* dj is Julian date -2415020 */
	t=(d/36525.0) - 1.0;		/* Now find t0 for this day */
	r0=t*(0.0513366 + t*(2.586222e-5 - t*1.722e-9));
	r1=6.697374558 + 2400.0*(t - ((e->year - 2000.0)/100.0));
	t0=r0 + r1;
	while (t0<0.0)			/* Get into range (0, 24) */
	    t0+=24.0;
	while (t0>24.0)
	    t0-=24.0;

	/* Now find UT (*put), Universal time by removing time zones & DST */
	*put = time - (double)ploc->dst - (double)ploc->time_zone;
	flg=0;
	while (*put<0.0)		/* Now get ut into (0,24) */
	    {
	    *put += 24.0;
	    flg=-1;
	    }
	while (*put>24.0)
	    {
	    *put-=24.0;
	    flg=1;
	    }
	t=t0;
	if (flg==-1)  t-=0.0657098244;  /* Sidereal corr per day */
	if (flg== 1)  t+=0.0657098244;

	gst = *put * 1.002737908 + t;
	while (gst<0.0)   gst+=24.0;
	while (gst>24.0)  gst-=24.0;

	return gst;
	}

/* cnvhadec2azel() (hour angle, declination) to (azimuth, elevation).
 * Structures of type hadec and type location carry the ins and a
 * structure of type azel the outs.  All angles
 * are in radians.
 */
void cnvhadec2azel
    (struct hadec *phd, struct location *ploc, struct azel *pae)
	{
	double a, b, sf, sx, sy, q;
	sx=sin(phd->ha);
	sy=sin(phd->dec);
	sf=sin(ploc->lat);
	a=sy*sf + cos(phd->ha)*cos(phd->dec)*cos(ploc->lat);
	q=asin(a);
	pae->el=q;
	b=cos(q)*cos(ploc->lat);
	if (fabs(b) < 1e-25)  b=1e-25;
	pae->az=acos((sy-a*sf)/b);
	if (sx>0.0)  pae->az=TPI-pae->az;
	}

/* cnvazel2hadec()
 * (azimuth, elevation) to (hour angle, declination).  Structures
 * of type azel and location carry the ins and a hadec the outs.  All angles
 * are in radians.
 */
void cnvazel2hadec
    (struct azel *pae, struct location *ploc, struct hadec *phd)
	{
	double a, b, sf, sx, sy, q;
	sx=sin(pae->az);
	sy=sin(pae->el);
	sf=sin(ploc->lat);
	a=sy*sf + cos(pae->az)*cos(pae->el)*cos(ploc->lat);
	q=asin(a);
	phd->dec=q;
	b=cos(q)*cos(ploc->lat);
	if (b<1e-10)  b=1e-10;
	phd->ha=acos((sy-a*sf)/b);
	if (sx>0.0)  phd->ha=TPI-phd->ha;
	}

/* cnvha2ra(double ha, double gst, double longitude)
 * Converts ha=Hour Angle into ra=Right Ascension.
 * gst=Greenwich Sidereal time (decimal hours)
 * lon=gegraphical longitude in radians (west is <0)
 * Returns Right Ascension in radians
 *
 * Same formulas convert ra to ha, and so cnvra2ha is a defined synonym.
 */
double cnvha2ra(double ha, double gst, double lon)
    {
    double lst, ra;

    lst=gst + r2h(lon);
    while (lst<0.0)   lst+=24.0;
    while (lst>=24.0) lst-=24.0;
    ra=lst - r2h(ha);
    while (ra<0.0)   ra+=24.0;
    while (ra>=24.0) ra-=24.0;
    return h2r(ra);
    }
    
/* obliq(eenv*) - Finds the angle between the Earths equator and the plane
 * of the ecliptic.  Called the obliquity of the ecliptic, it changes
 * slowly over the years.  Input is Julian date since 0.5,1,1900, dj.
 * Includes the nutation from nutat(), pe->ddo.
 */
void obliq(struct eenv *pe)
	{
	double t, a;
	t = (pe->dj / 36525.0) - 1.0;
	a = (46.815 + (0.0006 - 0.00181*t)*t)*t;
	a = d2r(a/3600.0);
	pe->ob = d2r(23.43929167) - a + pe->ddo;
	}

/* nutat(struct eenv *e) - Corrections to the earths ecliptic longitude
 * and obliquity of the ecliptic due to the gravity of the Moon and Sun.
 * Entered into structure for the Earth enviornment for a particular day
 * of type eenv.
 */
void nutat(struct eenv *e)
	{
	double lon_sr;		/* Mean longitude of Sun in radians */
	double lon_mr;		/* Mean longitude of Moon in radians */
	double anom_sr;		/* Mean anomaly of Sun in radians */
	double anom_mr;		/* Mean anomaly of Moon in radians */
	double lon_manr;	/* Lon of Moon's ascending node, radians */
	double lon_2manr;	/* 2*Lon of Moon's ascending node, radians */
	double a, b, t, t2;	/* Temporary variables */
	double dp, ddo;

	t = e->dj / 36525.0;
	t2 = t*t;
	a=100.0021358*t;
	b=360.0*fract_part(a);
	lon_sr=2.0*d2r(279.6967 + .000303*t2 + b);
	a=1336.855231*t;
	b=360.0*fract_part(a);
	lon_mr=2.0*d2r(270.4342 - .001133*t2 + b);

	a=99.99736056*t;
	b=360.0*fract_part(a);
	anom_sr=d2r(358.4758 - .00015*t2 + b);

	a=1325.552359*t;
	b=360.0*fract_part(a);
	anom_mr=d2r(296.1046 + .009192*t2 + b);

	a=5.372616667*t;
	b=360.0*fract_part(a);
	lon_manr=d2r(259.1833 + .002078*t2 - b);
	lon_2manr=2*lon_manr;

	dp=(-17.2327 - .01737*t)*sin(lon_manr)
	  + (-1.2729 - .00013*t)*sin(lon_sr) + .2088*sin(lon_2manr)
	  - .2037*sin(lon_mr) + (.1261 - .00031*t)*sin(anom_sr)
	  + .0675*sin(anom_mr) - (.0497 - .00012*t)*sin(lon_sr + anom_sr)
	  - .0342*sin(lon_mr - lon_manr) - .0261*sin(lon_mr + anom_mr)
	  + .0214*sin(lon_sr - anom_sr) - .0149*sin(lon_sr - lon_mr + anom_mr)
	  + .0124*sin(lon_sr - lon_manr) + .0114*sin(lon_mr - anom_mr);


	ddo=(9.21 + .00091*t)*cos(lon_manr)
	  + (.5522 - .00029*t)*cos(lon_sr) - .0904*cos(lon_2manr)
	  + .0884*cos(lon_mr) + .0216*cos(lon_sr + anom_sr)
	  + .0183*cos(lon_mr - lon_manr) + .0113*cos(lon_mr + anom_mr)
	  - .0093*cos(lon_sr - anom_sr) - .0066*cos(lon_sr - lon_manr);

	e->dp=d2r(dp/3600.0);
	e->ddo=d2r(ddo/3600.0);
	}

/* cnvll2radec()  -  Converts geocentric ecliptic lon and lat
 * to right ascension and declination.  We don't use it here, but
 * a slight change allows this function to go from RA & dec to lon & lat.
 * All angles in radians:
 *   *pll is lat, lon
 *   *prd is RA, dec
 *   pe->ob is obliquity of the ecliptic
 */
void cnvll2radec(struct latlon *pll, struct radec *prd, struct eenv *pe)
	{
	double o, co, so, cy, sy, cx, sx, p;	/* Local variables */

	so=sin(pe->ob);  co=cos(pe->ob);
	sy=sin(pll->lat);  cy=cos(pll->lat);
	if (fabs(cy) < 1.0e-20)  cy=1.0e-20;
	sx=sin(pll->lon);  cx=cos(pll->lon);
	if (fabs(cx) < 1.0e-20)  cx=1.0e-20;
	prd->dec = asin(sy*co + cy*so*sx);
	p=atan(((sx*co) - (sy*so/cy))/cx);
	if (cx<0)  p=p+PI;			/* Gives 4 quadrant ans */
	while (p>TPI)  p -= TPI;		/* Make (0, TPI) */
	while (p<0.0)  p += TPI;
	prd->ra = p;
	}

/* parallax( )  True to apparent conv, to correct for being on surface
 * of the Earth instead of at the center.
 * Height calculation is here as well, but at fixed 100 meter value.
 * Note: 6378140 meters is mean Earth radius.
 *
 *   lat=geographical latitude, radians
 *   HP=equatorial horizontal parallax, radians
 *   x =true hour angle, radians
 *   y =true declination, radians
*/
void parallax
    (double lat, struct hadec *hadect, double hp, struct hadec *hadeca)
    /*  in lat^            in x,y^       in hp^   out: aprnt ha,dec^    */
	{
	#define HEIGHT 1.56785521e-5            /* 100.0/6378140.0 */
	double x, y, rs, rc, rp, clat, slat, cx, cy, sy;  /* Local var */
	double p, q;

	x = hadect->ha;
	y = hadect->dec;

	clat=cos(lat);  slat=sin(lat);
	q=atan(0.996647*slat/clat);
	rs=0.996647*sin(q) + HEIGHT*slat;
	rc=cos(q) + HEIGHT*clat;
	rp=1/sin(hp);

	cx=cos(x);
	sy=sin(y);   cy=cos(y);
	p=x + atan(rc*sin(x)/(rp*cy - rc*cx));
	while (p>TPI)  p -= TPI;
	while (p<0.0)  p += TPI;
	q=atan(cos(p)*(rp*sy - rs)/(rp*cy*cx - rc));

	hadeca->ha  = p;
	hadeca->dec = q;
	}

/* anomaly(double am, ec, *at, *ae)  -  Returns
 * at, true anomaly and the ae, eccentric anomaly  in elliptical motion,
 * based on the mean anomaly am and the eccentricity, ec.
 * Iterative, returns the number of iterations, or 0 if excessive.
 * Anomalies are in radians, eccentricity is (0, 1.0).
 */
int anomaly(double am, double ec, double *pat, double *pae)
    {
    double m, d, a;
    int i;

    m = am - TPI*(floor)(am / TPI);
    *pae = m;
    d = -ec * sin(*pae);
    i = 0;
    while(fabs(d)>0.000000001 && i<100)
	{
	d = d / (1.0 - (ec * cos(*pae)));
	*pae -= d;
	d = *pae - ec * sin(*pae) - m;
	i++;
	}
    a=sqrt((1.0 + ec)/(1.0 - ec))*tan(*pae / 2.0);
    *pat = 2.0*atan(a);
    if (i>=99)
	return 0;
    else
	return i;
    }

/* void sun(double dj, double ut, double *sl, double *sed)
 * Finds sl, the true geocentric longitude of the Sun for the mean equinox
 * of the date and sed, the Sun-Earth distance in AU.
 * Inputs are dj, the modified Julian day from julday(), and ut, the
 * current Universal Time in decimal hours.
 */
void sun(double dj, double ut, double *psl, double *psed)
    {
    double a, b;
    double a1s, b1s, c1s, d1s, d2s, d3s, e1s;
    double h1s, ls, m1s, t, t2;
    double sam, sec, sat, sae;

    t=(dj/36525.0) + (ut/876600.0);
    t2=t*t;

    a=100.0021359*t;   b=360.0*fract_part(a);
    ls=279.69668 + 0.0003025*t2 + b;
    a=99.99736042*t;   b=360.0*fract_part(a);
    m1s=358.47583 - (0.00015 + 0.0000033*t)*t2 + b;
    sec = 1.675104E-02 - 0.0000418*t - 1.26E-07*t2;
    sam = d2r(m1s);
    anomaly(sam, sec, &sat, &sae);
    a=62.55209472*t;   b=360.0*fract_part(a);
    a1s=d2r(153.23 + b);
    a=125.1041894*t;   b=360.0*fract_part(a);
    b1s=d2r(216.57 + b);
    a=91.56766028*t;   b=360.0*fract_part(a);
    c1s=d2r(312.69 + b);
    a=1236.853095*t;   b=360.0*fract_part(a);
    d1s=d2r(350.74 - .00144*t2 + b);
    e1s=d2r(231.19 + 20.2*t);
    a=183.1353208*t;   b=360.0*fract_part(a);
    h1s=d2r(353.4 + b);

    d2s=0.00134*cos(a1s) + 0.00154*cos(b1s) + 0.002*cos(c1s)
       + 0.00179*sin(d1s) + 0.00178*sin(e1s);
    d3s=5.43E-06*sin(a1s) + 1.575E-05*sin(b1s)
       + 1.627e-05*sin(c1s) + 3.076e-05*cos(d1s) + 9.27e-06*sin(h1s);

    *psl=sat + d2r(ls - m1s + d2s);
    *psed=1.0000002*(1.0 - sec*cos(sae)) + d3s;
    while (*psl < 0)   *psl += TPI;
    while (*psl > TPI) *psl -= TPI;
    }

/* moon(dj, *pmm, *pbm, *ppm)
 *
 *  Inputs  dj=modified Julian day from julday()
 *  Global  ut=current Universal Time in decimal hours
 *  Outputs *pmm=true geocentric ecliptic longitude in radians
 *	    *pbm=true geocentric ecliptic latitude in radians
 *	    *pmrange=range c-c of E-M in meters
 */
void moon(double dj, double *pmm, double *pbm, double *ppm)
  {
  #define CM1 27.32158213
  #define CM2 365.2596407
  #define CM3 27.55455094
  #define CM4 29.53058868
  #define CM5 27.21222039
  #define CM6 6798.363307

  double t, t2, q;
  double mma, mmb, mmc, mmd, mme, mmf;
  double mln;				/* Mean Longitude */
  double mas;				/* Sun's mean anomaly */
  double mam;				/* Moon's mean anomoly */
  double mem;				/* Moon's mean elongation */
  double mdan;				/* Mean dist from ascending node */
  double lnan;				/* Long of ascending node */
  double mm;
  double mp;
  double a, b, c;
  double s1, s2, s3, s4;
  double e;
  double e2;
  double mlon;
  double mlat;
  double w1, w2;

  t = (dj/36525.0) + (ut/876600.0);	/* In centuries! */
  t2 = t*t;

  q = dj + (ut/24.0);
  mma=q/CM1;
  mmb=q/CM2;
  mmc=q/CM3;
  mmd=q/CM4;
  mme=q/CM5;
  mmf=q/CM6;
  mma=360*fract_part(mma);  mmb=360*fract_part(mmb);
  mmc=360*fract_part(mmc);  mmd=360*fract_part(mmd);
  mme=360*fract_part(mme);  mmf=360*fract_part(mmf);
  mln=270.434164 + mma - (0.001133 - 0.0000019*t)*t2;
  mas=358.475833 + mmb - (0.00015 + 0.0000033*t)*t2;
  mam=296.104608 + mmc + (0.009192 + 0.0000144*t)*t2;
  mem=350.737486 + mmd - (0.001436 - 0.0000019*t)*t2;
  mdan=11.250889 + mme - (0.003211 + 0.0000003*t)*t2;
  lnan=259.183275 - mmf + (0.002078 + 0.0000022*t)*t2;
  a=d2r(51.2 + 20.2*t);   s1=sin(a);  s2=sin(d2r(lnan));
  b=346.56 + (132.87 - 0.0091731*t)*t;
  s3=0.003964*sin(d2r(b));
  c=d2r(lnan + 275.05 - 2.3*t);
  s4=sin(c);
  mln=mln + 0.000233*s1 + s3 + 0.001964*s2;
  mas=mas - 0.001778*s1;
  mam=mam + 0.000817*s1 + s3 + 0.002541*s2;
  mdan=mdan + s3 - 0.024691*s2 - 0.004328*s4;
  mem=mem + 0.002011*s1 + s3 + 0.001964*s2;
  e=1 - (.002495 + 7.52E-06*t)*t;
  e2=e*e;
  mln=d2r(mln);  mas=d2r(mas);  lnan=d2r(lnan);
  mem=d2r(mem);  mdan=d2r(mdan);  mam=d2r(mam);

  mlon = 6.28875*sin(mam) + 1.274018*sin(2*mem - mam)
     + .658309*sin(2*mem) + .213616*sin(2*mam)
     - e*.185596*sin(mas) - .114336*sin(2*mdan)
     + 0.058793*sin(2*(mem - mam))
     + 0.057212*e*sin(2*mem - mas - mam) + 0.05332*sin(2*mem + mam)
     + 0.045874*e*sin(2*mem - mas) + 0.041024*e*sin(mam - mas)
     - 0.034718*sin(mem) - e*.030465*sin(mas + mam)
     + 0.015326*sin(2*(mem - mdan)) - 0.012528*sin(2*mdan + mam)
     - 0.01098*sin(2*mdan - mam) + 0.010674*sin(4*mem - mam)
     + 0.010034*sin(3*mam) + 0.008548*sin(4*mem - 2*mam)
     - e*.00791*sin(mas - mam + 2*mem) - e*.006783*sin(2*mem + mas)
     + 0.005162*sin(mam - mem) + e*.005*sin(mas + mem)
     + 0.003862*sin(4*mem) + e*.004049*sin(mam - mas + 2*mem)
     + 0.003996*sin(2*(mam + mem)) + 0.003665*sin(2*mem - 3*mam)
     + e*.002695*sin(2*mam - mas) + 0.002602*sin(mam - 2*(mdan + mem))
     + e*.002396*sin(2*(mem - mam) - mas) - 0.002349*sin(mam + mem)
     + e2*.002249*sin(2*(mem - mas)) - e*.002125*sin(2*mam + mas)
     - e2*.002079*sin(2*mas) + e2*.002059*sin(2*(mem - mas) - mam)
     - 0.001773*sin(mam + 2*(mem - mdan)) - 0.001595*sin(2*(mdan + mem))
     + e*.00122*sin(4*mem - mas - mam) - 0.00111*sin(2*(mam + mdan))
     + 0.000892*sin(mam - 3*mem) - e*.000811*sin(mas + mam + 2*mem)
     + e*.000761*sin(4*mem - mas - 2*mam)
     + e2*.000704*sin(mam - 2*(mas + mem))
     + e*.000693*sin(mas - 2*(mam - mem))
     + e*.000598*sin(2*(mem - mdan) - mas)
     + 0.00055*sin(mam + 4*mem) + 0.000538*sin(4*mam)
     + e*.000521*sin(4*mem - mas) + 0.000486*sin(2*mam - mem)
     + e2*.000717*sin(mam - 2*mas);
  mm=mln + d2r(mlon);
  while (mm < 0)  mm += TPI;
  while (mm > TPI)  mm -= TPI;
  *pmm = mm;

  mlat = 5.128189*sin(mdan) + 0.280606*sin(mam + mdan)
     + 0.277693*sin(mam - mdan) + 0.173238*sin(2*mem - mdan)
     + 0.055413*sin(2*mem + mdan - mam) + 0.046272*sin(2*mem - mdan - mam)
     + 0.032573*sin(2*mem + mdan) + 0.017198*sin(2*mam + mdan)
     + 0.009267*sin(2*mem + mam - mdan) + 0.008823*sin(2*mam - mdan)
     + e*.008247*sin(2*mem - mas - mdan) + 0.004323*sin(2*(mem - mam) - mdan)
     + 0.0042*sin(2*mem + mdan + mam) + e*0.003372*sin(mdan - mas - 2*mem)
     + e*0.002472*sin(2*mem + mdan - mas - mam)
     + e*0.002222*sin(2*mem + mdan - mas)
     + e*0.002072*sin(2*mem - mdan - mas - mam)
     + e*0.001877*sin(mdan - mas + mam) + 0.001828*sin(4*mem - mdan - mam)
     - e*0.001803*sin(mdan + mas) - 0.00175*sin(3*mdan)
     + e*0.00157*sin(mam - mas - mdan) - 0.001487*sin(mdan + mem)
     - e*0.001481*sin(mdan + mas + mam) + e*0.001417*sin(mdan - mas - mam)
     + e*0.00135*sin(mdan - mas) + 0.00133*sin(mdan - mem)
     + 0.001106*sin(mdan + 3*mam) + 0.00102*sin(4*mem - mdan)
     + 0.000833*sin(mdan + 4*mem - mam) + 0.000781*sin(mam - 3*mdan)
     + 0.00067*sin(mdan + 4*mem - 2*mam) + 0.000606*sin(2*mem - 3*mdan)
     + 0.000597*sin(2*(mem + mam) - mdan)
     + e*.000492*sin(2*mem + mam - mas - mdan)
     + 0.00045*sin(2*(mam - mem) - mdan)
     + 0.000439*sin(3*mam - mdan) + 0.000423*sin(mdan + 2*(mem + mam))
     + 0.000422*sin(2*mem - mdan - 3*mam)
     - e*0.000367*sin(mas + mdan + 2*mem - mam)
     - e*0.000353*sin(mas + mdan + 2*mem) + 0.000331*sin(mdan + 4*mem)
     + e*0.000317*sin(2*mem + mdan - mas + mam)
     + e2*0.000306*sin(2*(mem - mas) - mdan) - 0.000283*sin(mam + 3*mdan);
  w1=0.0004664*cos(lnan);
  w2=0.0000754*cos(c);
  *pbm = d2r(mlat)*(1.0 - w1 - w2);

  mp = 0.950724 + 0.051818*cos(mam) + 0.009531*cos(2*mem - mam)
    + 0.007843*cos(2*mem) + 0.002824*cos(2*mam)
    + 0.000857*cos(2*mem + mam) + e*0.000533*cos(2*mem - mas)
    + e*0.000401*cos(2*mem - mam - mas)
    + e*0.00032*cos(mam - mas) - 0.000271*cos(mem)
    - e*0.000264*cos(mas + mam) - 0.000198*cos(2*mdan - mam)
    + 0.000173*cos(3*mam) + 0.000167*cos(4*mem - mam)
    - e*0.000111*cos(mas) + 0.000103*cos(4*mem - 2*mam)
    - 0.000084*cos(2*mam - 2*mem) - e*0.000083*cos(2*mem + mas)
    + 0.000079*cos(2*mem + 2*mam) + 0.000072*cos(4*mem)
    + e*0.000064*cos(2*mem - mas + mam) - e*0.000063*cos(2*mem + mas - mam)
    + e*0.000041*cos(mas + mem) + e*0.000035*cos(2*mam - mas)
    - 0.000033*cos(3*mam - 2*mem) - 0.00003*cos(mam + mem)
    - 0.000029*cos(2*(mdan - mem)) - e*0.000029*cos(2*mam + mas)
    + e2*0.000026*cos(2*(mem - mas)) - 0.000023*cos(2*(mdan - mem) + mam)
    + e*0.000019*cos(4*mem - mas - mam);
  *ppm = d2r(mp);
  }
